
use std::os::raw::c_uint;

#[unsafe(no_mangle)]
pub static rsvg_major_version: c_uint = 2;

#[unsafe(no_mangle)]
pub static rsvg_minor_version: c_uint = 62;

#[unsafe(no_mangle)]
pub static rsvg_micro_version: c_uint = 2;
